#!/usr/bin/python
#
# Tools for generating reports

import urllib, sys, os.path
import sqlite3 as dbapi2
from subprocess import Popen
from cgi import escape

valid_states = ['inprogress','blocked','todo','done', 'postponed']
blueprints_base_url = 'https://blueprints.launchpad.net'

class user_string(str):
    pass

class team_string(str):
    pass

def escape_sql(text):
    return text.replace("'", "''")

def report_args(args, team=None, user=None, milestone=None):
    if milestone:
        args.extend(['-m',milestone])
    if team:
        args.extend(['-t',team])
    if user:
        args.extend(['-u',user])

def run_reports(my_path, database, basename, config, milestone=None, team=None,
        user=None, trend_starts=None, trend_override=None, burnup=False):

    ident = (team or user)
    if milestone and ident:
        chartname = ident + '-' + milestone
    elif milestone:
        chartname = 'all-' + milestone
    elif ident:
        chartname = ident
    else:
        chartname = 'all' 

    cfg = load_config(config)

    hf = open(basename + '.html', 'w')
    args = [os.path.join(my_path, 'html-report'), '-d', database, '--chart', '%s.svg' % chartname]
    if 'title' in cfg:
        args += ['--title', cfg['title']]
    report_args(args, team, user, milestone)
    hreport = Popen(args, stdout=hf)

    jf = open(basename + '.json', 'w')
    args = [os.path.join(my_path, 'json-report'), '-d', database, '-c', config]
    report_args(args, team, user, milestone)
    jreport = Popen(args, stdout=jf)

    args = [os.path.join(my_path, 'burndown-chart'), '-d', database, '-o', basename + '.svg']
    report_args(args, team, user, milestone)

    if trend_override:
        args.extend(['--trend-start', trend_override])
    else:
        if milestone and ident:
            if (ident, milestone) in trend_starts:
                args.extend(['--trend-start', str(trend_starts[(ident, milestone)])])
        elif ident:
            if ident in trend_starts:
                args.extend(['--trend-start', str(trend_starts[ident])])
    if burnup:
        args.extend(['--only-weekdays', '--inverted'])
    burndown = Popen(args)

    print basename + '.html'
    hreport.wait()
    hf.close()
    print basename + '.json'
    jreport.wait()
    jf.close()
    print basename + '.svg'
    burndown.wait()


def fix_stdouterr():
    '''Set stdout/stderr encoding.

    Avoid having to do .encode('UTF-8') everywhere. This is a pain; I wish
    Python supported something like "sys.stdout.encoding = 'UTF-8'".
    '''
    import codecs
    def null_decode(input, errors='strict'):
        return input, len(input)
    sys.stdout = codecs.EncodedFile(sys.stdout, 'UTF-8')
    sys.stdout.decode = null_decode
    sys.stderr = codecs.EncodedFile(sys.stderr, 'UTF-8')
    sys.stderr.decode = null_decode

def get_db(dbpath):
    '''Open database.'''

    assert os.path.exists(dbpath)
    return dbapi2.connect(dbpath)

def load_config(config_path):
    '''Load a configuration file and return it as a dictionary.'''

    cfg = {}
    execfile(config_path, cfg)
    del cfg['__builtins__']
    return cfg

def escape_url(url):
    '''Correctly escape data intended to be included in a URL'''

    return urllib.quote(url)

def escape_html(html):
    '''Correctly escape data intended to be included as raw HTML'''

    return escape(html, True)

def workitems_over_time(db, team=None, milestone=None):
    '''Calculate work item development over time.

    If team is given, this filters out blueprints for that particular team. If
    team_assignees_only is True, only work items for that team's members are
    counted, otherwise this counts all work items for specs belonging to the
    given team.

    Return date -> state -> count mapping. states are
    {todo,inprogress,done,postponed}{,_teamonly}.
    '''
    data = {}
    if milestone:
        # get milestone date range
        cur = db.cursor()
        cur.execute('SELECT due_date FROM milestones WHERE name = ?', (milestone,))
        date_end = cur.fetchone()[0]
        cur.execute('SELECT MAX(due_date), MAX(due_date) < date(CURRENT_TIMESTAMP) FROM milestones WHERE due_date < ?', (date_end,))
        (date_start, in_milestone) = cur.fetchone()

        # if we are already within the milestone, start at the previous
        # milestone's end date; if we are before, start at the beginning of the
        # entire cycle, to allow planning
        if in_milestone:
            ms_sql = "AND w.milestone='%s' AND date >= '%s' AND date <= '%s'" % (
                    escape_sql(milestone), date_start, date_end)
        else:
            ms_sql = "AND w.milestone='%s' AND date <= '%s'" % (
                    escape_sql(milestone), date_end)
    else:
        ms_sql = ''

    # include both direct team assignments, as well as assignmets to
    # members of that team
    cur = db.cursor()
    if team:
        # WIs which are assigned to team members
        if isinstance(team, user_string):
            cur.execute(
                    'SELECT status, date, count(status) FROM ('
                    'SELECT DISTINCT date, w.spec, w.description, w.assignee, w.milestone, w.status AS status '
                    'FROM work_items w '
                    'WHERE assignee = ? %s '
                    ') GROUP BY status, date' % ms_sql, (team,))
        else:
            cur.execute(
                    'SELECT status, date, count(status) FROM ('
                    'SELECT DISTINCT date, w.spec, w.description, w.assignee, w.milestone, w.status AS status '
                    'FROM work_items w LEFT JOIN teams '
                    'ON w.assignee = teams.name '
                    'WHERE (team = ? OR assignee = ?) %s'
                    ') GROUP BY status, date' % ms_sql, (team, team))

        for (s, date, num) in cur:
            data.setdefault(date, {})[s + '_teamonly'] = num

        # all WIs which belong to that team's specs
        # Only do this for teams, users only want their
        # assigned work items
        if isinstance(team, user_string):
            cur.execute(
                    'SELECT status, date, count(status) FROM ('
                    'SELECT DISTINCT date, w.spec, w.description, w.assignee, w.milestone, w.status AS status '
                    'FROM work_items w, specs s ON w.spec = s.name '
                    'WHERE (w.assignee = ? OR (s.assignee = ? and w.assignee is null)) %s'
                    ') GROUP BY status, date' % ms_sql, (team, team))
        else:
            cur.execute(
                    'SELECT status, date, count(status) FROM ('
                    'SELECT DISTINCT date, w.spec, w.description, w.assignee, w.milestone, w.status AS status '
                    'FROM work_items w, specs s ON w.spec = s.name '
                    '   LEFT JOIN teams ON (s.assignee = teams.name OR w.assignee = teams.name) '
                    'WHERE (teams.team = ? OR w.assignee = ? OR s.assignee = ?) %s'
                    ') GROUP BY status, date' % ms_sql, (team, team, team))

        for (s, date, num) in cur:
            data.setdefault(date, {})[s] = num

    else:
        # all WIs for this milestone
        cur.execute('SELECT status, date, count(*) FROM work_items w '
                'WHERE 1=1 %s GROUP BY status, date' % ms_sql)

        for (s, date, num) in cur:
            data.setdefault(date, {})[s] = num
    return data

def blueprint_completion(db, team=None, milestone=None):
    '''Determine current blueprint completion.

    Return blueprint -> info mapping, with info being a map with these
    keys: todo, inprogress, done, postponed, status, priority, implementation, url.
    '''
    data = {}
    if milestone == '':
        ms_sql = "AND w.milestone IS NULL"
    elif milestone:
        ms_sql = "AND w.milestone='%s'" % escape_sql(milestone)
    else:
        ms_sql = ''

    # last date
    cur = db.cursor()
    cur.execute('SELECT max(date) FROM work_items')
    (last_date,) = cur.fetchone()

    cur = db.cursor()
    if team:
        # include both direct team assignments, as well as assignmnets to
        # members of that team
        if isinstance(team, user_string):
            cur.execute('SELECT w.status, s.name, COUNT(DISTINCT w.description), s.status, s.priority, s.implementation, s.url, s.roadmap_notes, s.assignee '
                    'FROM work_items w, specs s ON w.spec = s.name '
                    'WHERE w.date = ? AND'
                    '  (w.assignee = ? or (s.assignee = ? and w.assignee is null)) %s '
                    'GROUP BY w.status,w.spec' % ms_sql,
                    (last_date, team, team))
        else:
            cur.execute('SELECT w.status, s.name, COUNT(DISTINCT w.description), s.status, s.priority, s.implementation, s.url, s.roadmap_notes, s.assignee '
                    'FROM work_items w, specs s ON w.spec = s.name '
                    '   LEFT JOIN teams ON (s.assignee = teams.name OR w.assignee = teams.name)'
                    'WHERE w.date = ? AND'
                    '  (teams.team = ? OR w.assignee = ? or s.assignee = ?) %s '
                    'GROUP BY w.status, w.spec' % ms_sql,
                    (last_date, team, team, team))
    else:
        cur.execute('SELECT w.status, s.name, COUNT(DISTINCT w.description), s.status, s.priority, s.implementation, s.url, s.roadmap_notes, s.assignee '
                'FROM work_items w, specs s '
                'ON w.spec = s.name '
                'WHERE w.date = ? %s '
                'GROUP BY w.status, w.spec' % ms_sql,
                (last_date,))

    for (s, bp, num, status, priority, impl, url, roadmap_notes, assignee) in cur:
        info = data.setdefault(bp, {'todo': 0, 'blocked': 0, 'done': 0, 'postponed': 0, 'inprogress': 0})
        info[s] = num
        info['status'] = status or ''
        info['priority'] = priority
        info['implementation'] = impl
        info['url'] = url
        info['roadmap_notes'] = roadmap_notes
        info['assignee'] = assignee
        
        cur2 = db.cursor();
        cur2.execute('SELECT SUM(points) FROM complexity as w WHERE spec = ? AND date = ? %s' % ms_sql, (bp, last_date))
        for points in cur2:
            info['complexity'] = points[0] 

    return data


def blueprint_tasks_completion(db, team=None, milestone=None):
    '''Determine current blueprint tasks completion.

    Return blueprint -> info mapping, with info being a map with these
    keys: todo, blocked, inprogress, done, postponed, status, priority, implementation, url, tasks.
    '''
    default = blueprint_completion(db, team=team)
    data = {}
    if milestone == '':
        ms_sql = "AND w.milestone IS NULL"
    elif milestone:
        ms_sql = "AND w.milestone='%s'" % escape_sql(milestone)
    else:
        ms_sql = ''

    # last date
    cur = db.cursor()
    cur.execute('SELECT max(date) FROM work_items')
    (last_date,) = cur.fetchone()

    for s in valid_states:
        cur = db.cursor()
        if team:
            # include both direct team assignments, as well as assignmnets to
            # members of that team
            if isinstance(team, user_string):
                cur.execute('SELECT DISTINCT w.spec, w.description, w.assignee, w.milestone, w.status '
                        'FROM work_items w, specs s ON w.spec = s.name '
                        'WHERE w.status = ? AND w.date = ? AND'
                        '  (w.assignee = ? or (s.assignee = ? and w.assignee is null)) %s ' % ms_sql,
                        (s, last_date, team, team))
            else:
                cur.execute('SELECT DISTINCT w.spec, w.description, w.assignee, w.milestone, w.status '
                        'FROM work_items w, specs s ON w.spec = s.name '
                        '   LEFT JOIN teams ON (s.assignee = teams.name OR w.assignee = teams.name) '
                        'WHERE w.status = ? AND w.date = ? AND'
                        '  (teams.team = ? OR w.assignee = ? or s.assignee = ?) %s ' % ms_sql,
                        (s, last_date, team, team, team))
        else:
            cur.execute('SELECT DISTINCT w.spec, w.description, w.assignee, w.milestone, w.status '
                    'FROM work_items w, specs s '
                    'ON w.spec = s.name '
                    'WHERE w.status = ? AND w.date = ? %s '
                    'GROUP BY w.spec' % ms_sql,
                    (s, last_date))

        for (bp, desc, assignee, milestone, status) in cur:
            data.setdefault(bp, default[bp])
            info = data[bp].setdefault('tasks', [])
            info.append({
                'description': desc,
                'assignee': assignee,
                'milestone': milestone,
                'state': status
            })

    return data


def assignee_completion(db, team=None, milestone=None):
    '''Determine current by-assignee completion.

    Return assignee -> info mapping with info being a map with these
    keys: todo, blocked, inprogress, done, postponed. Each of those values is a list of [blueprint,
    workitem, priority, spec_url].
    '''
    data = {}
    if milestone == '':
        ms_sql = "AND w.milestone IS NULL"
    elif milestone:
        ms_sql = "AND w.milestone='%s'" % milestone
    else:
        ms_sql = ''

    # last date
    cur = db.cursor()
    cur.execute('SELECT max(date) FROM work_items')
    (last_date,) = cur.fetchone()

    for s in valid_states:
        cur = db.cursor()
        if team:
            # include both direct team assignments, as well as assignmnets to
            # members of that team
            if isinstance(team, user_string):
                cur.execute('SELECT DISTINCT w.assignee, w.spec, w.description, s.priority, s.url '
                        'FROM work_items w, specs s ON w.spec = s.name '
                        'WHERE w.status = ? AND w.date = ? AND'
                        '  (w.assignee = ? or (s.assignee = ? and w.assignee is null)) %s ' % ms_sql,
                        (s, last_date, team, team))
            else:
                cur.execute('SELECT DISTINCT w.assignee, w.spec, w.description, s.priority, s.url '
                        'FROM work_items w, specs s ON w.spec = s.name '
                        '   LEFT JOIN teams ON (s.assignee = teams.name OR w.assignee = teams.name) '
                        'WHERE w.status = ? AND w.date = ? AND'
                        '  (teams.team = ? OR w.assignee = ? or s.assignee = ?) %s ' % ms_sql,
                        (s, last_date, team, team, team))
        else:
            cur.execute('SELECT DISTINCT w.assignee, w.spec, w.description, s.priority, s.url '
                    'FROM work_items w, specs s '
                    'ON w.spec = s.name '
                    'WHERE w.status = ? AND w.date = ? %s ' % ms_sql,
                    (s, last_date))
        for (a, bp, description, priority, url) in cur:
            info = data.setdefault(a, {'todo': [], 'blocked': [], 'done': [], 'postponed': [], 'inprogress': []})
            info[s].append([bp, description, priority, url])

            cur2 = db.cursor();
            cur2.execute('SELECT SUM(points) FROM complexity as w WHERE assignee = ? AND date = ? %s' % ms_sql, (a, last_date))
            for points in cur2:
                info['complexity'] = points[0]

    return data

def spec_information( db, config, team=None, milestone=None ):
    '''The meta information a spec exposes through Launchpad

    Returns a dictionary of specs -> meta information closely
    mirroring the DB structure
    '''

    ### cargo culted from assignee_completion
    if milestone == '':
        ms_sql = "AND milestone IS NULL"
    elif milestone:
        ms_sql = "AND milestone='%s'" % milestone
    else:
        ms_sql = ''

    ### limit by a team 'like'? (see config)
    if team:
        lookup = ''
        try:
            lookup = config.get('teams')[team]
        except KeyError:
            lookup = team
        team_sql = "AND name like '%%%s%%' " % lookup
    else:
        team_sql = ""

    completion_dict = blueprint_completion( db, team, milestone )

    # last date
    cur = db.cursor()
    cur.execute('SELECT max(date) FROM work_items')
    (last_date,) = cur.fetchone()

    ### need team information to map assignee => team
    teams = {}
    t_cur = db.cursor()
    t_cur.execute('SELECT name,team FROM teams')
    for n,t in t_cur.fetchall():
        teams[n] = t

    sql = "select * from specs where 1=1 " + ms_sql + " " + team_sql

    db.row_factory = dbapi2.Row
    cur = db.cursor()
    cur.execute(sql)

    rv = {}


    ### surely.. there's a better way?
    for spec_row in cur.fetchall():

        spec_dict = {
            "work_items":   [],
            "meta":         [],
            "completion":   {},
        }

        ### map the result into key/value pairs
        # .keys is new in 2.6, so this blows up the tracker...
        # must follow this instead:
        # http://www.python.org/dev/peps/pep-0249/
        #for k in spec_row.keys():
        #    spec_dict[k] = spec_row[k]
        for d in cur.description:
            spec_dict[ d[0] ] = spec_row[ d[0] ]

        ### the 'team' column is empty in the specs table,
        ### so we'll find the info ourselves

        ### XXX have to do this *after* the spec_row.keys lookup,
        ### or it will return None, blowing up the dictionary get
        spec_dict["team"] = teams.get( spec_row['assignee' ],
                              teams.get( spec_row['approver' ], None )
                            )

        ### use the todo/done/postponed numbers from blueprint_completion
        csd = completion_dict.get( spec_row["name"], None )
        if csd:
            for k in valid_states:
                spec_dict["completion"][k] = csd[k]


        ### extend it with the work item data
        wi_cur = db.cursor()
        wi_cur.execute( "select * from work_items where date = ? and spec = ? "
            + ms_sql, (last_date, spec_row["name"],) )

        for wi_row in wi_cur.fetchall():
            wi_dict = {}

            ### map the result into key/value pairs -- 2.6'ism :(
            #for k in wi_row.keys():
            #    wi_dict[k] = wi_row[k]

            for d in wi_cur.description:
                wi_dict[ d[0] ] = wi_row[ d[0] ]

            spec_dict[ "work_items" ].append( wi_dict )

        ### extend it with the meta information
        meta_cur = db.cursor()
        meta_cur.execute( "select * from meta where date = ? and spec = ?",
                          ( last_date, spec_row["name"] ) )

        for meta_row in meta_cur.fetchall():
            meta_dict = {}

            # 2.6'ism :(
            #for k in meta_row.keys():
            #    meta_dict[k] = meta_row[k]

            for d in meta_cur.description:
                meta_dict[ d[0] ] = meta_row[ d[0] ]

            spec_dict["meta"].append( meta_dict )

        rv[ spec_row["name"] ] = spec_dict

    return rv

def team_information( db, config, team=None ):
    '''The team information as returned by LP

    Returns a dictionary with teamname -> members
    '''

    if team:
        teams = [ team ]
    else:
        cur = db.cursor()
        cur.execute('SELECT DISTINCT team FROM teams')
        teams = [i[0] for i in cur]

    rv = {}

    for t in teams:
        cur = db.cursor()
        cur.execute( 'SELECT name FROM teams WHERE team = ?', (t,) )

        rv[t] = [ i[0] for i in cur ]

    return rv


def milestone_list(db):
        milestones = []

        cur = db.cursor()
        cur.execute('SELECT m.name FROM milestones m ORDER BY m.due_date')

        for (milestone,) in cur:
            milestones.append(milestone)

        return milestones


def priority_value(name):
    '''Translate a priority string into a sortable numeric value.'''

    trans = {'Essential': 99, 'High': 75, 'Medium': 50, 'Low': 25}

    return trans.get(name, 0)


def fill_template(name, data):
    from mako.lookup import TemplateLookup
    lookup = TemplateLookup(directories=['templates'])
    template = lookup.get_template(name)
    return template.render_unicode(**data)


if __name__ == '__main__':
    # some test stuff
    db = get_db(sys.argv[1])

    print '==== Assignee completion ==='
    for date, data in assignee_completion(db).iteritems():
        print date, [len(data[x]) for x in data]
    print '--------------'
    for date, data in assignee_completion(db, milestone='lucid-alpha-2').iteritems():
        print date, [len(data[x]) for x in data]
    print '--------------'
    for date, data in assignee_completion(db, 'canonical-desktop-team').iteritems():
        print date, [len(data[x]) for x in data]
    print '--------------'
    for date, data in assignee_completion(db, 'canonical-desktop-team', milestone='lucid-alpha-2').iteritems():
        print date, [len(data[x]) for x in data]
